#include <stdio.h>
#include <CL/cl.h>

int main() {

   #define N 256
   size_t size = N * N * sizeof(float);

   cl_event ev_writeA, ev_writeB, ev_ker, ev_readC;
   int i;

   float * matA = malloc(N * N * sizeof(float));
   float * matB = malloc(N * N * sizeof(float));
   float * matC = malloc(N * N * sizeof(float));

   for (i=0; i<N*N; i++) {
      matA[i] = (float)i;
      matB[i] = (float)i + 1.0;
   }

   cl_uint nb_platf;
   clGetPlatformIDs(0, NULL, &nb_platf);

   printf("Nombre de plateformes: %d\n", nb_platf);

   cl_platform_id platfs[nb_platf];
   clGetPlatformIDs(nb_platf, platfs, NULL);

   size_t plat_name_size;
   clGetPlatformInfo(platfs[0], CL_PLATFORM_NAME, 0, NULL, &plat_name_size);
   char plat_name[plat_name_size];
   clGetPlatformInfo(platfs[0], CL_PLATFORM_NAME, plat_name_size, &plat_name, NULL);
   printf("Nom de la plateforme: %s\n", plat_name);

   size_t plat_vendor_size;
   clGetPlatformInfo(platfs[0], CL_PLATFORM_VENDOR, 0, NULL, &plat_vendor_size);
   char plat_vendor[plat_vendor_size];
   clGetPlatformInfo(platfs[0], CL_PLATFORM_VENDOR, plat_vendor_size, &plat_vendor, NULL);
   printf("Nom de la plateforme: %s\n", plat_vendor);

   cl_uint nb_devs;
   clGetDeviceIDs(platfs[0], CL_DEVICE_TYPE_ALL, 0, NULL, &nb_devs);

   cl_device_id devs[nb_devs];
   clGetDeviceIDs(platfs[0], CL_DEVICE_TYPE_ALL, nb_devs, devs, NULL);

   cl_context ctx = clCreateContext(NULL, nb_devs, devs, NULL, NULL, NULL);
   cl_command_queue cq = clCreateCommandQueue(ctx, devs[0], CL_QUEUE_OUT_OF_ORDER_EXEC_MODE_ENABLE | CL_QUEUE_PROFILING_ENABLE, NULL);

   FILE * f = fopen("./matrix_add.cl", "r");
   fseek(f, 0, SEEK_END);
   size_t source_size = ftell(f);
   fseek(f, 0, SEEK_SET);

   char * source = malloc(source_size+1);
   fread(source, 1, source_size, f);
   source[source_size] = '\0';
   fclose(f);

   printf("Code kernel: %s\n", source);

   cl_program prg = clCreateProgramWithSource(ctx, 1, (const char**)&source, NULL, NULL);
   clBuildProgram(prg, nb_devs, devs, NULL, NULL, NULL);

   size_t log_size;
   clGetProgramBuildInfo(prg, devs[0], CL_PROGRAM_BUILD_LOG, 0, NULL, &log_size);
   char log[log_size];
   clGetProgramBuildInfo(prg, devs[0], CL_PROGRAM_BUILD_LOG, log_size, log, NULL);

   printf("Build log: %s\n", log);

   cl_kernel ker = clCreateKernel(prg, "matrix_add", NULL);

   cl_mem bufA = clCreateBuffer(ctx, CL_MEM_READ_ONLY, size, NULL, NULL);
   cl_mem bufB = clCreateBuffer(ctx, CL_MEM_READ_ONLY, size, NULL, NULL);
   cl_mem bufC = clCreateBuffer(ctx, CL_MEM_WRITE_ONLY, size, NULL, NULL);

   clEnqueueWriteBuffer(cq, bufA, 0, 0, size, matA, 0, NULL, &ev_writeA);
   clEnqueueWriteBuffer(cq, bufB, 0, 0, size, matB, 0, NULL, &ev_writeB);

   clSetKernelArg(ker, 0, sizeof(bufA), &bufA);
   clSetKernelArg(ker, 1, sizeof(bufB), &bufB);
   clSetKernelArg(ker, 2, sizeof(bufC), &bufC);

   size_t globalDim[] = {N, N};
   size_t localDim[] = {256, 1};
   cl_event deps[] = {ev_writeA, ev_writeB};
   clEnqueueNDRangeKernel(cq, ker, 2, NULL, globalDim, localDim, 2, deps, &ev_ker);

   clEnqueueReadBuffer(cq, bufC, 0, 0, size, matC, 1, &ev_ker, &ev_readC);

   clFinish(cq);

   for (i=0; i<N; i++) {
      int j;
      for (j=0; j<N; j++) {
         printf("%f\t", matC[i*N + j]);
      }
      printf("\n");
   }

   printf("Fini.\n");

   return 0;
}
